"""
Day 29 - Bouncing Balls

Balls of slowly changing colors, bouncing over the leds.
Unlike the previous effects, this should work reasonably for both 2D and 3D layouts.
Works best for fairly dense layouts though.
A continuous effect. You can upload it as a movie instead if you prefer, but it will jerk in the transition.
Remember to adjust aspect ratio in the call to adjust_layout_aspect in the code below.
"""

from xled_plus.samples.sample_setup import *

class MeanderBouncingScene(BouncingScene):
    def __init__(self, ctr, num, speed=0.25, size=0.3):
        super(MeanderBouncingScene, self).__init__(ctr, num, speed, size, False)
        self.bgcol = hsl_color(0.0, 0.0, -0.5)
        self.bgeffect = BreathEffect(ctr, [(0.0, 0.0, 0.0)], 1.0, 0.75, [12, 30])
        self.colfunc = False
        self.horizon = 0
        self.preferred_frames = 640

    def create(self):
        sh = super(MeanderBouncingScene, self).create()
        sh.cm = ColorMeander("surface", start=(random(), 1.0, 0.0))
        return sh

    def update(self, step):
        for sh in self.shapes:
            sh.cm.step()
            (h, s, l) = sh.cm.get_hsl()
            sh.color = hsl_color(h, 1.0, min(0.3, max(-0.3, l)))
        super(MeanderBouncingScene, self).update(step)

    def blend_colors(self, colors):
        cols = colors
        br = list(map(lambda rgb: color_brightness(*rgb), colors))
        maxbr = max(br) if br else 0.0
        sumbr = sum(br) if br else 0.0
        bgbr = color_brightness(*self.bgcol)
        if maxbr < bgbr:
            delta = (bgbr - maxbr) / bgbr
            sumbr += (bgbr - maxbr)
            maxbr = bgbr
            cols = cols + [tuple(map(lambda x: x*delta, self.bgcol))]
        sumcol = tuple(map(lambda *args: max(0, min(255, int(round(sum(args) * maxbr / sumbr)))), *cols))
        return sumcol

    def reset(self, numframes):
        super(MeanderBouncingScene, self).reset(numframes)
        self.bgeffect.reset(numframes)

    def getnext(self):
        pat1 = super(MeanderBouncingScene, self).getnext()
        patbg = self.bgeffect.getnext()
        vec = self.getoccupancy()
        pat = [pat1[i] if vec[i] else patbg[i] for i in range(self.ctr.num_leds)]
        return pat


if __name__ == '__main__':
    ctr = setup_control()
    ctr.adjust_layout_aspect(1.0)  # How many times wider than high is the led installation?
    eff = MeanderBouncingScene(ctr, 5, size=0.4)
    oldmode = ctr.get_mode()["mode"]
    eff.launch_rt()
    print("Started continuous effect - press Return to stop it")
    input()
    eff.stop_rt()
    ctr.set_mode(oldmode)



"""
xled_plus.sequence
~~~~~~~~~~~~~~~~~~

Author: Anders Holst (anders.holst@ri.se), 2021

This module contains some classes for creating spatio-temporal cyclic color
sequences. That is, bands or gradients of colors moving across the leds in
some direction at some speed.

It assumes that a 2D or 3D layout is set on the leds, and that any adjustment
to the aspect ratio has been made with HighControlInterface.adjust_layout_aspect
to make the angles come out right. In the 2D case the 'angle' parameter below
is just a single angle, in degreed counted clockwise from the top. In the 3D
case, 'angle' is a tuple (theta, phi), where theta is the polar angle, and phi
the azimuthal angle.
"""


from xled_plus.effect_base import Effect
from xled_plus.colormeander import ColorMeander
from xled_plus.pattern import blendcolors, dimcolor
from xled_plus.ledcolor import hsl_color
import math as m


class Sequence(Effect):
    def __init__(self, ctr, seqfunc, speed, folds, angle=False):
        super(Sequence, self).__init__(ctr)
        self.seqfunc = seqfunc
        if not ctr.layout_bounds:
            ctr.fetch_layout()
        self.dim = ctr.layout_bounds["dim"]
        if self.dim > 1:
            theta = (
                (angle[0] if isinstance(angle, tuple) else angle or 0.0) * m.pi / 180.0
            )
            phi = (angle[1] if isinstance(angle, tuple) else 0.0) * m.pi / 180.0
            dy = m.cos(theta) * folds / 2.0
            dxz = m.sin(theta) * folds / 2.0
            if self.dim == 3:
                self.vect = (dxz * m.cos(phi), dy, dxz * m.sin(phi))
            else:
                self.vect = (dxz, dy)
        else:
            self.vect = (folds / 2.0,)
        self.speed = speed
        self.currpos = 0.0
        self.init_fps(self.preferred_fps)

    def init_fps(self, fps):
        self.preferred_fps = fps
        if self.speed % fps == 0:
            self.preferred_frames = 1
        else:
            cyclen = 1.0 / abs((self.speed / fps + 0.5) % 1.0 - 0.5)
            self.preferred_frames = int(round(cyclen))

    def set_vector(self, vec):
        assert len(vec) == self.dim
        self.vect = vec

    def dot(self, v1, v2):
        return sum(map(lambda x1, x2: x1 * x2, v1, v2))

    def reset(self, numframes):
        self.currpos = 0.0

    def update(self, step):
        self.currpos += self.speed * step

    def getnext(self):
        self.update(1.0 / self.preferred_fps)
        return self.ctr.make_layout_pattern(
            lambda pos: self.seqfunc((self.dot(self.vect, pos) + self.currpos) % 1.0),
            style="centered",
        )


class ColorSequence(Sequence):
    def __init__(self, ctr, cols, lens=False, speed=1.0, folds=1.0, angle=False):
        super(ColorSequence, self).__init__(ctr, self.getcolor, speed, folds, angle)
        if lens is False:
            self.lims = [(i + 1) / len(cols) for i in range(len(cols))]
        else:
            self.lims = []
            acc = 0.0
            tot = float(sum(lens))
            for i in range(len(cols)):
                acc += lens[i] / tot
                self.lims.append(acc)
        self.cols = [c for c in cols]

    def lookup(self, x, lims):
        for i in range(len(lims)):
            if x < lims[i]:
                return i
        return False

    def getcolor(self, x):
        return self.cols[self.lookup(x, self.lims)]


class GradientSequence(Sequence):
    def __init__(self, ctr, cols, lens=False, speed=1.0, folds=1.0, angle=0):
        super(GradientSequence, self).__init__(ctr, self.getcolor, speed, folds, angle)
        if lens is False:
            self.lims = [float(i) / len(cols) for i in range(len(cols) + 1)]
        else:
            self.lims = [0.0]
            acc = 0.0
            tot = float(sum(lens))
            for i in range(len(cols)):
                acc += lens[i] / tot
                self.lims.append(acc)
        self.lens = lens
        self.cols = [c for c in cols]
        self.cols.append(cols[0])

    def lookup(self, x, lims):
        for i in range(len(lims)):
            if x < lims[i]:
                return i
        return False

    def getcolor(self, x):
        ind = self.lookup(x, self.lims)
        return blendcolors(
            self.cols[ind - 1],
            self.cols[ind],
            (x - self.lims[ind - 1]) / (self.lims[ind] - self.lims[ind - 1]),
        )


class SpectrumSequence(Sequence):
    def __init__(self, ctr, lightness=0.0, angle=False):
        super(SpectrumSequence, self).__init__(ctr, self.getcolor, 0.2, 1.0, angle)
        self.lightness = lightness

    def getcolor(self, x):
        return hsl_color(x, 1.0, self.lightness)


class RotatingAngleSequence(Sequence):
    def initialize(self, torque):
        deltaang = m.pi / 180.0 * torque / self.preferred_fps
        self.sa = m.sin(deltaang)
        self.ca = m.cos(deltaang)
        if self.dim == 1:
            self.ivect = (self.vect[0], 0.0)

    def update(self, step):
        if self.dim == 3:
            self.vect = (
                self.vect[0] * self.ca + self.vect[2] * self.sa,
                self.vect[1],
                self.vect[2] * self.ca - self.vect[0] * self.sa,
            )
        elif self.dim == 2:
            self.vect = (
                self.vect[0] * self.ca + self.vect[1] * self.sa,
                self.vect[1] * self.ca - self.vect[0] * self.sa,
            )
        else:
            self.ivect = (
                self.ivect[0] * self.ca + self.ivect[1] * self.sa,
                self.ivect[1] * self.ca - self.ivect[0] * self.sa,
            )
            self.vect = (self.ivect[0], )
        self.currpos += self.speed * step


class RotatingAngleColorSequence(ColorSequence, RotatingAngleSequence):
    def __init__(self, ctr, cols, lens=False, speed=1.0, folds=1.0, angle=False, torque=False):
        super(RotatingAngleColorSequence, self).__init__(ctr, cols, speed=speed, folds=folds, angle=angle)
        self.initialize(torque)


class RotatingAngleGradientSequence(GradientSequence, RotatingAngleSequence):
    def __init__(self, ctr, cols, lens=False, speed=1.0, folds=1.0, angle=False, torque=False):
        super(RotatingAngleGradientSequence, self).__init__(ctr, cols, speed=speed, folds=folds, angle=angle)
        self.initialize(torque)


class VaryingAngleSequence(Sequence):
    def initialize(self, maxfold):
        self.maxfold = maxfold
        if self.dim == 3:
            self.mangle = ColorMeander("sphere")
        else:
            self.mangle = ColorMeander("cylinder")

    def update(self, step):
        self.mangle.step()
        (x, y, z) = self.mangle.get_xyz()
        if self.dim == 3:
            self.vect = (
                x * self.maxfold / 2.0,
                y * self.maxfold / 2.0,
                z * self.maxfold / 2.0,
            )
        elif self.dim == 2:
            self.vect = (x * self.maxfold / 2.0, y * self.maxfold / 2.0)
        else:
            self.vect = (z * self.maxfold / 2.0,)
        self.currpos += self.speed * step


class VaryingAngleColorSequence(ColorSequence, VaryingAngleSequence):
    def __init__(self, ctr, cols, lens=False, speed=1.0, folds=3.0):
        super(VaryingAngleColorSequence, self).__init__(ctr, cols, speed, folds)
        self.initialize(folds)


class VaryingAngleGradientSequence(GradientSequence, VaryingAngleSequence):
    def __init__(self, ctr, cols, lens=False, speed=1.0, folds=3.0):
        super(VaryingAngleGradientSequence, self).__init__(ctr, cols, speed, folds)
        self.initialize(folds)


class MeanderingSequence(VaryingAngleSequence):
    def __init__(self, ctr, fixangle=False):
        super(MeanderingSequence, self).__init__(ctr, self.getcolor, 0.0, 1.0, fixangle)
        self.fixangle = fixangle
        self.initialize(1.0)
        self.cm = ColorMeander("sphere")
        self.colvec = [self.cm.get()] * 200
        self.iter = 1
        self.init_fps(2)
        self.preferred_frames = 800

    def reset(self, numframes):
        self.currpos = 0.5
        if numframes:
            self.mangle.steplen *= 10
            self.mangle.noiselev *= 3
            self.iter = 9
            self.preferred_fps /= 10

    def update(self, step):
        if self.fixangle is False:
            super(MeanderingSequence, self).update(step)
        for i in range(self.iter):
            self.cm.step()
            self.colvec = [self.cm.get()] + self.colvec[:-1]

    def getcolor(self, x):
        ind = int(x * len(self.colvec) - 0.5)
        return self.colvec[min(max(0, ind), len(self.colvec)-1)]


class MeanderingTandemSequence(VaryingAngleSequence):
    def __init__(self, ctr, gap=False, fixangle=False):
        super(MeanderingTandemSequence, self).__init__(ctr, self.getcolor, 0.0, 1.0, fixangle)
        self.initialize(1.0)
        self.gap = gap
        self.fixangle = fixangle
        self.cm = ColorMeander("sphere")
        self.init_fps(2)
        self.preferred_frames = 800

    def reset(self, numframes):
        self.currpos = 0.5
        if numframes:
            self.cm.steplen *= 10
            self.cm.noiselev *= 3
            self.preferred_fps /= 10

    def update(self, step):
        if self.fixangle is False:
            super(MeanderingTandemSequence, self).update(step)
        self.cm.step()
        (h, s, l) = self.cm.get_hsl()
        self.hsl1 = (h, s, l)
        self.hsl2 = ((h + 0.5) % 1.0, s, l)

    def getcolor(self, x):
        hsl = self.hsl1 if x < 0.5 else self.hsl2
        fact = min(1.0, abs(2*x - 1.0) / abs(self.gap)) if self.gap else 1.0
        if self.gap >= 0.0:
            return hsl_color(hsl[0], hsl[1]*fact, hsl[2])
        else:
            return hsl_color(hsl[0], hsl[1], hsl[2]) if fact == 1.0 else (0,0,0)




__version__ = '0.1.33'

_classifiers = [
    'Development Status :: 4 - Beta',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Programming Language :: Python :: 3.10',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: MIT License',
    'Topic :: Software Development :: Libraries',
    'Topic :: Utilities',
]

readme = """
XLED Plus provides addons to the XLED package. Whereas XLED provides a
python interface to Twinkly led lights, XLED Plus provides classes and
functions to make it easier to produce various effects on the lights
entirely from python: still or animated, simple or advanced,
prerecorded or created in real time.
"""

def _run_setup():
    from setuptools import setup

    setup(
        name='xled_plus',
        version=__version__,
        author='Anders Holst',
        author_email='anders.holst@ri.se',
        url='https://github.com/Anders-Holst/xled_plus',
        packages=['xled_plus','xled_plus.samples'],
        description='Addons to the XLED package for controling Twinkly LED lights',
        long_description=readme,
        license='MIT',
        classifiers=_classifiers,
        keywords=['xled','twinkly','light-effects'],
        install_requires=['xled'],
    )


if __name__ == '__main__':
    _run_setup()



"""
Day 19 - Rotating Star

A rotating star on a color shifting background (and another star
behind it rotating the other way).
"""

from xled_plus.samples.sample_setup import *

class RotatingStarScene(Scene):
    def __init__(self, ctr):
        super(RotatingStarScene, self).__init__(ctr)
        self.star1c = Star(5, (0, 0), 0.0, 0.25, 0.1, hsl_color(0.38, 1.0, 0.4))
        self.star1b = Star(5, (0, 0), 0.0, 0.5, 0.2, hsl_color(0.40, 1.0, 0.2))
        self.star1a = Star(5, (0, 0), 0.0, 0.75, 0.3, hsl_color(0.42, 1.0, 0.0))
        self.star2 = Star(5, (0, 0), 0.0, 0.75, 0.4, hsl_color(0.48, 1.0, 0.1))
        self.star1a.set_torque(24 / self.preferred_fps)
        self.star1b.set_torque(24 / self.preferred_fps)
        self.star1c.set_torque(24 / self.preferred_fps)
        self.star2.set_torque(-24 / self.preferred_fps)
        self.star1a.set_depth(1.2)
        self.star1b.set_depth(1.1)
        self.star1c.set_depth(1.0)
        self.star2.set_depth(2)
        self.add_shape(self.star1a)
        self.add_shape(self.star1b)
        self.add_shape(self.star1c)
        self.add_shape(self.star2)
        self.proj2D3D = "cylbase"

    def update(self, step):
        self.star1a.update(step)
        self.star1b.update(step)
        self.star1c.update(step)
        self.star2.update(step)

    def reset(self, numframes):
        super(RotatingStarScene, self).reset(numframes)
        self.time = 0

    def getnext(self):
        pat1 = super(RotatingStarScene, self).getnext()
        hue = (0.875 + 0.5 * abs(0.5 - (float(self.time) / self.preferred_frames))) % 1.0
        patbg = ctr.make_layout_pattern(lambda pos: hsl_color(hue, 1.0, max(0.0, 0.9 - (pos[0]**2+pos[1]**2))), style="centered")
        vec = self.getoccupancy()
        pat = [pat1[i] if vec[i] else patbg[i] for i in range(self.ctr.num_leds)]
        self.time += 1
        return pat

ctr = setup_control()
ctr.adjust_layout_aspect(1.0)  # How many times wider than high is the led installation?
RotatingStarScene(ctr).launch_movie()



"""
Day 3 - Gold

Similar to the Sparkles effect in the app, of bright flashing leds
against a solid background. This example is meant to mimic the
"metallic luster" effects of AWW leds, i.e glittering gold.
"""

from xled_plus.samples.sample_setup import *

ctr = setup_control()
Gold(ctr).launch_movie()



"""
Day 5 - Dark to light colors

Gradient from dark to light of a slowly changing hue across the spectrum.

This effect uses the physical layout of the leds. It only uses
one coordinate though (y-axis), so it should work for 2D, 3D, or
linear layouts. (Older firmware do not store the layout on the leds.
For them the effect will follow the string instead.)
"""

from xled_plus.samples.sample_setup import *

def darklight(t, pos):
    light = pos[1] if len(pos) > 1 else pos[0]
    hue = t % 1.0
    return hsl_color(hue, 1.0, light*2 - 1)

ctr = setup_control()
mov = ctr.make_func_movie(120, lambda t: ctr.make_layout_pattern(lambda pos: darklight(t / 120.0, pos)))
ctr.show_movie(mov, 0.66)



"""
Day 21 - Moon

It is winter solstice - the longest night of the year.
And just a couple of days ago was a full moon.
This effect focuses on the night and the phases of the moon.
In it, the moon passes all of its phases in just a minute,
on a background of sparkling stars.

Again, gives best effect on a fairly dense 2D layout, and adjust
the aspect ratio below if you need.
"""

from xled_plus.samples.sample_setup import *

class Moon(Ellipse):
    def __init__(self, cent, rad, light, dark):
        super(Moon, self).__init__(cent, 0.0, rad, rad, None)
        self.light = light
        self.dark = dark
        self.r2 = rad*rad
        self.tilt = 0.5
        self.phase = -180.0
        self.speed = 1.0

    def moonlit(self, pos):
        ca = m.cos(m.pi * self.phase * self.tilt / 360.0)
        sa = m.sin(m.pi * self.phase * self.tilt / 360.0)
        cb = abs(m.cos(m.pi * (self.phase % 180.0) / 180.0))
        y = -ca * pos[1] + sa * pos[0]
        x = (-ca * pos[0] - sa * pos[1]) / cb if cb != 0.0 else 0.0
        if self.phase < -90:
            return x < 0.0 and x**2 + y**2 > self.r2
        elif self.phase == -90:
            return x < 0.0
        elif self.phase < 0:
            return x < 0.0 or x**2 + y**2 < self.r2
        elif self.phase < 90:
            return x > 0.0 or x**2 + y**2 < self.r2
        elif self.phase == 90:
            return x > 0.0
        else:
            return x > 0.0 and x**2 + y**2 > self.r2
        
    def update(self, step):
        self.phase = (self.phase + 180.0 + self.speed * step) % 360 - 180.0

    def get_color(self, coord, ind):
        if self.is_inside(coord):
            if self.moonlit(coord):
                return self.light
            else:
                return self.dark
        else:
            return False

class MoonScene(MovingShapesScene):
    def __init__(self, ctr):
        super(MoonScene, self).__init__(ctr)
        self.bgeff = SparkleEffect(ctr, 3, self.whitefunc, pulselight_func(10, 1, 10))
        self.moon = Moon((0, 0), 0.4, hsl_color(0, 0, 1), (1, 1, 1))
        self.add_shape(self.moon)
        self.preferred_frames = 360
        self.preferred_fps = 6
        self.proj2D3D = "cylshell"

    def whitefunc(self, *args):
        r = random()
        return hsl_color(0.05 if r < 0.5 else 0.45, 1.0, 0.8 - abs(r - 0.5))

    def update(self, step):
        self.moon.update(step)
        
    def reset(self, numframes):
        self.bgeff.reset(numframes)

    def getnext(self):
        pat1 = super(MoonScene, self).getnext()
        pat2 = self.bgeff.getnext()
        vec = self.getoccupancy()
        pat = [pat1[i] if vec[i] else pat2[i] for i in range(self.ctr.num_leds)]
        return pat

ctr = setup_control()
ctr.adjust_layout_aspect(1.0)  # How many times wider than high is the led installation?
MoonScene(ctr).launch_movie()



"""
xled_plus.colormeander
~~~~~~~~~~~~~~~~~

Author: Anders Holst (anders.holst@ri.se), 2021

Performing a random walk through color space

The idea with this module is to produce a sequence of slowly changing colors,
not just cycling through the spectrum in a deterministic and repetitive way,
but through the entire 3D color space, never repeating exactly the same path.

To use it, create a ColorMeander object (no mandatory arguments, but optional
ones described below). By default it starts at full white. By calling 'step'
it moves one step in color space. You can then retrieve the current color
coordinates by 'get()' (for rgb coordinates), 'get_hsl()' (for hsl coordinates),
'get_xyz()' (for the raw x,y,z-coordinates if you want to interpret them as a
color in some other way), or 'get_compl()' which gives the rgb coordinates of
the complementary color (opposite hue but same lightness and saturation).

It can be customized by a number of optional arguments when created:

'style' can be 'sphere' (the default, moves inside spherical color coordinates),
'cylinder' (to move inside cylindrical coordinates) or 'surface' (just moving on
the surface of the color sphere).

'speed' is the length to move inside the color space, should be much less than 1
for best effect, default is 0.01.

'noise' is the level of the random noise to add to the current direction. Should
also be less than 1 for best effect. Default is 0.15.

'start' is the starting point in space. Default is (0,0,1) which corresponds
to white. (0,0,-1) means black. (0,0,0) means mid gray. The initial direction
is random.
"""

from math import asin, atan2, sqrt, pi
import random

from xled_plus.ledcolor import hsl_color


def hyp(*args):
    return sqrt(sum(map(lambda x: x * x, args)))


class ColorMeander:
    def __init__(self, style="sphere", speed=0.01, noise=0.15, start=(0.0, 0.0, 1.0)):
        self.steplen = speed
        self.noiselev = noise
        self.xyz = start
        self.dir = (
            random.random() - 0.5,
            random.random() - 0.5,
            random.random() - 0.5 - start[2],
        )
        self.style = style

    def normalize(self, vec):
        nrm = hyp(*vec)
        if nrm == 0.0:
            nrm = 1.0
        return tuple(map(lambda v: v / nrm, vec))

    def xyz_to_hsl(self, x, y, z):
        if self.style == "cylinder":
            h = atan2(y, x) / (2 * pi) + 0.5
            s = min(1.0, hyp(x, y))
            l = z
        else:
            h = atan2(y, x) / (2 * pi) + 0.5
            l = asin(z) * 2.0 / pi
            r = sqrt(x * x + y * y)
            r0 = sqrt(1 - z * z)
            s = min(1.0, r / r0 if r0 > 0.0 else 0.0)
        return (h, s, l)

    def xyz_color(self, x, y, z):
        return hsl_color(*self.xyz_to_hsl(x, y, z))

    def step(self):
        (nx, ny, nz) = tuple(map(lambda v, d: v + d * self.steplen, self.xyz, self.dir))
        if self.style == "cylinder":
            if abs(nz) > 1.0:
                nz = max(-1.0, min(1.0, nz))
            nrm = hyp(nx, ny)
            if nrm > 1.0:
                nx = nx / nrm
                ny = ny / nrm
                self.dir = self.normalize(
                    (nx - self.xyz[0], ny - self.xyz[1], nz - self.xyz[2])
                )
            ndir = self.normalize(
                tuple(
                    map(
                        lambda v: v
                        + random.random() * 2 * self.noiselev
                        - self.noiselev,
                        self.dir,
                    )
                )
            )
            if abs(nz + ndir[2]) > 1.0:
                sgn = 1 if nz + ndir[2] > 0.0 else -1
                delta = sqrt(1.0 - (sgn - nz) ** 2)
                nrm = hyp(ndir[0], ndir[1])
                ndir = (ndir[0] * delta / nrm, ndir[1] * delta / nrm, sgn - nz)
        elif self.style == "surface":
            nrm = hyp(nx, ny, nz)
            nx = nx / nrm
            ny = ny / nrm
            nz = nz / nrm
            ndir = self.normalize(
                (nx - self.xyz[0], ny - self.xyz[1], nz - self.xyz[2])
            )
            ndir = self.normalize(
                tuple(
                    map(
                        lambda v: v
                        + random.random() * 2 * self.noiselev
                        - self.noiselev,
                        ndir,
                    )
                )
            )
        else:
            nrm = hyp(nx, ny, nz)
            if nrm > 1.0:
                nrm = nrm * nrm  # bounce equally much inside
                nx = nx / nrm
                ny = ny / nrm
                nz = nz / nrm
                self.dir = self.normalize(
                    (nx - self.xyz[0], ny - self.xyz[1], nz - self.xyz[2])
                )
            ndir = self.normalize(
                tuple(
                    map(
                        lambda v: v
                        + random.random() * 2 * self.noiselev
                        - self.noiselev,
                        self.dir,
                    )
                )
            )
        self.xyz = (nx, ny, nz)
        self.dir = ndir

    def get(self):
        return self.xyz_color(*self.xyz)

    def get_compl(self):
        return self.xyz_color(-self.xyz[0], -self.xyz[1], self.xyz[2])

    def get_xyz(self):
        return self.xyz

    def get_hsl(self):
        return self.xyz_to_hsl(*self.xyz)



"""
xled_plus.ledcolor
~~~~~~~~~~~~~

Author: Anders Holst (anders.holst@ri.se), 2021

This module provides a color model adapted for the led lights.
It is based on HSL (Hue - Saturation - Lightness).

The main entry-point is hsl_color(h, s, l) which takes values for hue,
saturation and lightness, and returns a tuple of rgb values.
Hue is in the range 0.0 - 1.0 (where both ends represent blue).
Saturation is also in the range 0.0 - 1.0.
Lighness however is in the range -1.0 - +1.0, where -1.0 is black, +1.0
is white, and 0.0 represents the maximum chroma of the hue.

You can choose between 4 color circles and two lightness policies, with the
function set_color_style which understands the following tags:
"3col": Standard R-G-B color circle with the colors Blue-Cyan-Green-
        -Yelow-Red-Magenta at equal distances.
"4col": NCS-color circle with Blue-Green-Yellow-Red at equal distances.
        Not particularly useful here but included if someone prefers it.
"6col": The common paintbox color circle with Blue-Green-Yellow-Orange-
        -Red-Purple at equal distances
"8col": A combined color circle, with Blue-Cyan-Green-Yellow-Orange-Red-
        -Magenta-Purple at equal distances. This is the most "perceptually
        even" circle, and in addition happens to have complementary colors
        placed opposite to each other.
"linear": A traditional lightness policy, where the light increases
          linearly for each hue from black (-1) up to maximum chroma (0)
          and then linearly from there to white (1). However, different
          hues have then different lightness with this policy.
"equilight": A lightness policy enforcing that all hues at a given
             lightness have the same perceived brightness. Maximum
             chroma occurs at different lightness for different hues.
A standard HSL color model would have "3col" and "linear", whereas the
default setting here is "8col" and "equilight" because it is more intuitive
and easier to work with.

The RGB type led lights are originally not well white balanced. They
have a cold purplish tinge when fully lit. The provided color model
compensates for this by balancing the lights. With this model you
can easily select any white you want - neutral, warm, cold, pinkish
or greenish - without the need to go to RGBW.
"""

led_gamma = 1.0

led_brightness = [0.35, 0.50, 0.15]

led_balance = [0.9, 1.0, 0.6]

col_styles_dict = {
    "3col": [0.0, 1.0 / 6.0, 2.0 / 6.0, 3.0 / 6.0, 4.0 / 6.0, 5.0 / 6.0, 1.0],
    "4col": [0.0, 1.0 / 8.0, 1.0 / 4.0, 2.0 / 4.0, 3.0 / 4.0, 7.0 / 8.0, 1.0],
    "6col": [0.0, 1.0 / 12.0, 1.0 / 6.0, 1.0 / 3.0, 2.0 / 3.0, 3.0 / 4.0, 1.0],
    "8col": [0.0, 1.0 / 8.0, 2.0 / 8.0, 3.0 / 8.0, 5.0 / 8.0, 6.0 / 8.0, 1.0],
    "10col": [0.0, 2.0 / 10.0, 3.0 / 10.0, 4.0 / 10.0, 7.0 / 10.0, 8.0 / 10.0, 1.0],
}

col_style = ("8col", "equilight")


# Internal functions


def colorgamma(x):
    global led_gamma
    return x if led_gamma == 1.0 else pow(x, led_gamma)


def invcolorgamma(x):
    global led_gamma
    return x if led_gamma == 1.0 else pow(x, 1.0 / led_gamma)


def colorgamma_image(x):
    return (pow(x, 1.0 / 2.4) * 1.055 - 0.055) if x > 0.0031308 else x * 12.92


def invcolorgamma_image(x):
    return pow((x + 0.055) / 1.055, 2.4) if x > 0.04045 else x / 12.92


def color_brightness(r, g, b):
    return sum(map(lambda c, br: c * br, [r, g, b], led_brightness))


# Entry points


def set_color_style(style):
    """
    Set the color circle and lightness policy to use in the color model.
    Possible color circles are: "3col", "4col", "6col" and "8col".
    Possible lightness policies are: "linear" and "equilight".

    :param str style: color circle or lightness policy to use.
    :rtype: tuple
    """
    global col_style, col_styles_dict
    if style in ["linear", "equilight"]:
        col_style = (col_style[0], style)
        return col_style
    elif style in col_styles_dict:
        col_style = (style, col_style[1])
        return col_style
    else:
        return False


def get_color_style():
    """
    Return the currently used color circle and lightness policy as a tuple.

    :rtype: tuple
    """
    global col_style
    return col_style


def rgb_color(r, g, b):
    """
    Takes r, g and b values in the range 0.0 - 1.0, and converts it to an
    rgb tuple in the range 0-255, adjusted for white balance of the leds.

    :param float r: red component (0.0 - 1.0)
    :param float g: green component (0.0 - 1.0)
    :param float b: blue component (0.0 - 1.0)
    :rtype: tuple
    """
    global led_balance
    return tuple(
        map(
            lambda c, b: max(0, min(255, int(255 * b * colorgamma(c)))),
            [r, g, b],
            led_balance,
        )
    )


def image_to_led_rgb(r, g, b):
    """
    Converts rgb values for a computer image pixel into rgb values for a led,
    compensating for gamma correction in the image.
    The input values and the returned rgb values are all in the range 0-255.

    :param float r: red component (0 - 255)
    :param float g: green component (0 - 255)
    :param float b: blue component (0 - 255)
    :rtype: tuple
    """
    return tuple(
        map(
            lambda c, bal: max(
                0, min(255, int(255 * bal * colorgamma(invcolorgamma_image(c / 255.0))))
            ),
            [r, g, b],
            led_balance,
        )
    )


def led_to_image_rgb(r, g, b):
    """
    Converts rgb values for a led into rbg values for a computer image pixel,
    compensating for gamma correction in the image.
    The input values and the returned rgb values are all in the range 0-255.

    :param float r: red component (0 - 255)
    :param float g: green component (0 - 255)
    :param float b: blue component (0 - 255)
    :rtype: tuple
    """
    return tuple(
        map(
            lambda c, bal: max(
                0,
                min(255, int(255 * colorgamma_image(invcolorgamma(c / (bal * 255.0))))),
            ),
            [r, g, b],
            led_balance,
        )
    )


def hsl_color(h, s, l):
    """
    Takes hue (0.0 - 1.0), saturation (0.0 - 1.0), and lightness (-1.0 - 1.0)
    values and converts it to an rgb tuple in the range 0-255.

    :param float h: hue component (0.0 - 1.0)
    :param float s: saturation component (0.0 - 1.0)
    :param float l: lightness component (-1.0 - 1.0)
    :rtype: tuple
    """
    global col_style, col_styles_dict
    hramp = col_styles_dict[col_style[0]]
    ir = 1.0 / led_balance[0]
    ig = 1.0 / led_balance[1]
    ib = 1.0 / led_balance[2]
    irg = min(ir, ig)
    irb = min(ir, ib)
    igb = min(ig, ib)
    iramp = [
        (0, 0, ib),
        (0, igb / 2, igb / 2),
        (0, ig, 0),
        (irg / 2, irg / 2, 0),
        (ir, 0, 0),
        (irb / 2, 0, irb / 2),
        (0, 0, ib),
    ]
    i = 0
    while h > hramp[i + 1]:
        i += 1
    p = (h - hramp[i]) / (hramp[i + 1] - hramp[i])
    (r, g, b) = tuple(map(lambda x1, x2: p * (x2 - x1) + x1, iramp[i], iramp[i + 1]))
    nrm = max(r / ir, g / ig, b / ib)
    (r, g, b) = tuple(map(lambda x: x / nrm, (r, g, b)))
    ll = (l + 1.0) * 0.5
    if col_style[1] == "linear":
        if ll < 0.5:
            t1 = l + 1.0
            t2 = 0.0
        else:
            t1 = 1.0 - l
            t2 = l
    else:
        br = color_brightness(r, g, b)
        # make the hue get its maximum dynamic saturation, up till maximum green, then linearly decreasing
        e = max(r, g, b)
        p = min(
            1.0,
            (1.0 - ll / e) / (1.0 - br),
            (1.0 - ll * led_balance[1]) / (1.0 - led_brightness[1]),
        )
        t1 = ll * p / ((br - e) * p + e)
        t2 = max(0.0, ll - t1 * br)
    t1 = s * t1
    t2 = s * t2 + ll * (1.0 - s)
    return rgb_color(r * t1 + t2, g * t1 + t2, b * t1 + t2)



"""
Day 2 - Water

Bluish, greenish and white, like glittering water in a pond.
Similar to what can be achieved by the Glow effect in the app, but
colors change instead of just pulsing, and the transition to the first
frame is smooth, rather than with that annoying glitch you get with
the effect in the app.
"""

from xled_plus.samples.sample_setup import *

ctr = setup_control()
Water(ctr).launch_movie()
